import time
from typing import Dict

from notebuild.tool.fastapi import add_api_routes, api_route
from notecoin.database.base import create_all, create_session
from notecoin.database.connect import RedisConnect
from notecoin.okex.database.client import OkexClientAccountBalance
from notecoin.okex.database.strategy import OkexStrategyAutoSeller
from notecoin.okex.database.websocket import OkexSocketPublicTickers


class AutoSeller(RedisConnect):
    def __init__(self, prefix="/sell", *args, **kwargs):
        self.total = 0
        self.usdt = 0

        self.session = create_session()
        create_all()
        self.db_account: Dict[str, OkexClientAccountBalance] = {}
        self.db_seller: Dict[str, OkexStrategyAutoSeller] = {}

        super(AutoSeller, self).__init__(prefix=prefix, *args, **kwargs)
        add_api_routes(self)

    def load_account(self):
        self.db_account.clear()
        self.db_seller.clear()
        for detail in self.session.query(OkexClientAccountBalance).all():
            if detail.availBal > 1:
                self.db_account[detail.ccy] = detail
        for detail in self.session.query(OkexStrategyAutoSeller).all():
            self.db_seller[detail.coin_id] = detail
        return {}

    def update_price(self):
        details = self.session.query(OkexSocketPublicTickers).all()
        data_map = dict([(detail.instId, detail.last) for detail in details])

        for aa in self.db_account.values():
            coin_id = aa.ccy if aa.ccy == 'USDT' else f"{aa.ccy}-USDT"
            if coin_id not in self.db_seller.keys():
                tmp = OkexStrategyAutoSeller.instance(coin_id=coin_id,
                                                      worth=aa.eqUsd,
                                                      init_worth=50,
                                                      count=aa.availBal,
                                                      price=aa.eqUsd / aa.availBal)
                self.session.merge(tmp)
                self.session.commit()
            if coin_id not in self.db_seller.keys():
                continue
            price = data_map.get(coin_id, 1.0)
            coin_seller = self.db_seller[coin_id]
            coin_seller.update_worth(worth=price * aa.availBal, count=aa.availBal, price=price)
        return {
            # "price_size": len(data_map),
            # "price_data": data_map
        }

    def to_json(self):
        self.total = self.usdt
        for coin in self.db_seller.values():
            self.total += coin.worth

        return {
            "total": round(self.total, 2),
            "res": round(self.usdt, 2),
            "coins": [coin.to_json() for coin in self.db_seller.values()]
        }

    @api_route("/update")
    def update_value(self, suffix=""):
        res = {}
        res.update(self.load_account())
        res.update(self.update_price())

        for coin in self.db_seller.values():
            if coin.coin_id == 'USDT':
                self.usdt = coin.worth
                continue
            if coin.check():
                coin.sell()
                self.session.query(OkexStrategyAutoSeller).filter(
                    OkexStrategyAutoSeller.coin_id == coin.coin_id).delete()
            else:
                self.session.merge(coin)
            self.session.commit()
        update_time = int(round(time.time() * 1000)) - 60 * 60 * 1000
        self.session.query(OkexStrategyAutoSeller).filter(OkexStrategyAutoSeller.updateTime < update_time).delete()
        self.session.commit()
        res.update(self.to_json())
        return res



import csv
import logging
import os
import warnings

import pandas as pd
from notesecret.secret import read_secret
from sqlalchemy import MetaData, Table, create_engine, inspect, select
from sqlalchemy.orm import declarative_base
from tqdm import tqdm

meta = MetaData()

uri = read_secret(cate1='notecoin', cate2='dataset', cate3='db_path')
uri = uri or f'sqlite:///{os.path.abspath(os.path.dirname(__file__))}/data/notecoin.db'

engine = create_engine(uri)
Base = declarative_base()
logging.info(f'uri:{uri}')


def create_engines(url, **kwargs):
    return create_engine(url, **kwargs)


class _BaseTable:
    def __init__(self, table_name, engine, *args, **kwargs):
        self.table_name = table_name
        self.table: Table = None
        self.engine = engine

    def create(self):
        meta.create_all(self.engine)

    def insert(self, values, keys=None, *args, **kwargs):
        cols = [col.name for col in self.table.columns]
        if isinstance(values, dict):
            values = dict([(k, v) for k, v in values.items() if k in cols])
        elif isinstance(values, list):
            if isinstance(values[0], dict):
                values = [dict([(k, v) for k, v in item.items() if k in cols]) for item in values]
            elif isinstance(values[0], list):
                values = [dict(zip(keys, item)) for item in values]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # code here...
            if str(self.engine.url).startswith('sqlite'):
                ins = self.table.insert(values=values).prefix_with("OR IGNORE")
            else:
                ins = self.table.insert(values=values).prefix_with("IGNORE")
            self.engine.execute(ins)

    def select_all(self):
        return self.engine.execute(select([self.table]))

    def delete_all(self):
        return self.engine.execute(select([self.table]).delete())

    def to_csv(self, file_path, result=None):
        if result is None:
            result = self.select_all()
        with open(file_path, 'wb') as fw:
            out_csv = csv.writer(fw)
            out_csv.writerow(result.keys())
            out_csv.writerows(result)

    def to_csv_all(self, file_path, page_size=100000, total_step=1000):
        for step in tqdm(range(total_step)):
            start, stop = page_size * step, page_size * (step + 1)
            df = pd.read_sql(select([self.table]).slice(start, stop), self.engine)
            if len(df) == 0:
                break
            if start == 0:
                df.to_csv(file_path, header=True, index=False, mode='w')
            else:
                df.to_csv(file_path, header=False, index=False, mode='a')

    def upsert(self, values):
        if str(self.engine.url).startswith('mysql'):
            from sqlalchemy.dialects.mysql import insert
        elif str(self.engine.url).startswith('sqlite'):
            from sqlalchemy.dialects.sqlite import insert
        else:
            from sqlalchemy.dialects.postgresql import insert

        stmt = insert(self.table).values(values)
        primary_keys = [key.name for key in inspect(self.table).primary_key]
        update_dict = {c.name: c for c in stmt.excluded if
                       not c.primary_key and c.name not in ('gmt_create', 'gmt_creat')}
        stmt = stmt.on_conflict_do_update(index_elements=primary_keys, set_=update_dict)
        self.engine.execute(stmt)


class BaseTable(_BaseTable):
    def __init__(self, table_name, *args, **kwargs):
        super(BaseTable, self).__init__(table_name, engine=engine, *args, **kwargs)



import json
import logging
import time

import pandas as pd
from ccxt.base.exchange import Exchange
from notecoin.base.database import KlineData
from tqdm import tqdm

logger = logging.getLogger()
unix_month = 2678400000


class LoadDataKline:
    def __init__(self, exchange: Exchange, *args, **kwargs):
        self.table = KlineData(source=exchange.name)
        self.exchange = exchange
        super(LoadDataKline, self).__init__(*args, **kwargs)

    def load_all(self, *args, **kwargs):
        self.exchange.load_markets()
        for sym in tqdm(self.exchange.symbols):
            if ':' not in sym:
                self.load(sym, *args, **kwargs)

    def _load(self, symbol, fetch_since, timeframe='1m', limit=100, *args, **kwargs):
        result = self.exchange.fetch_ohlcv(symbol, timeframe, fetch_since, limit=limit)
        result = self.exchange.sort_by(result, 0)
        if len(result) == 0:
            return False, f'result is empty, fetch_since:{fetch_since}'

        df = pd.DataFrame(result, columns=['timestamp', 'open', 'close', 'low', 'high', 'vol'])
        df['symbol'] = symbol
        self.table.insert(json.loads(df.to_json(orient='records')))
        time.sleep(int(self.exchange.rateLimit / 1000))
        return True, result

    def load(self, symbol, timeframe='1m', limit=100, unix_start=None, unix_end=None, *args, **kwargs):
        unix_max, unix_min = self.table.select_symbol_maxmin(symbol)
        unix_delta = limit * self.exchange.parse_timeframe(timeframe) * 1000

        unix_start = unix_start or self.exchange.milliseconds() - unix_month
        unix_end = unix_end or self.exchange.milliseconds()
        unix_max = unix_max or self.exchange.milliseconds()
        unix_min = unix_min or self.exchange.milliseconds()

        unix_temp = unix_min
        pbar = tqdm(range(unix_min, unix_start, -unix_delta), desc=symbol)
        for _ in pbar:
            status, result = self._load(symbol, unix_temp - unix_delta, timeframe=timeframe, limit=limit)
            if status is False:
                logger.info("result is null")
                break
            pbar.set_postfix({'fetch_since': self.exchange.iso8601(result[0][0])})
            unix_temp = result[0][0]

        unix_temp = unix_max
        pbar = tqdm(range(unix_max, unix_end, unix_delta), desc=symbol)
        for _ in pbar:
            status, result = self._load(symbol, unix_temp, timeframe=timeframe, limit=limit)
            if status is False:
                print(result)
                break
            unix_temp = result[-1][0]



import time
import zlib


def get_local_timestamp():
    return int(time.time())


def partial(res):
    data_obj = res['data'][0]
    bids = data_obj['bids']
    asks = data_obj['asks']
    instrument_id = res['arg']['instId']
    # print('全量数据bids为:' + str(bids))
    # print('档数为:' + str(len(bids)))
    # print('全量数据asks为:' + str(asks))
    # print('档数为:' + str(len(asks)))
    return bids, asks, instrument_id


def update_bids(res, bids_p):
    # 获取增量bids数据
    bids_u = res['data'][0]['bids']
    # print('增量数据bids为:' + str(bids_u))
    # print('档数为:' + str(len(bids_u)))
    # bids合并
    for i in bids_u:
        bid_price = i[0]
        for j in bids_p:
            if bid_price == j[0]:
                if i[1] == '0':
                    bids_p.remove(j)
                    break
                else:
                    del j[1]
                    j.insert(1, i[1])
                    break
        else:
            if i[1] != "0":
                bids_p.append(i)
    else:
        bids_p.sort(key=lambda price: sort_num(price[0]), reverse=True)
        # print('合并后的bids为:' + str(bids_p) + '，档数为:' + str(len(bids_p)))
    return bids_p


def update_asks(res, asks_p):
    # 获取增量asks数据
    asks_u = res['data'][0]['asks']
    # print('增量数据asks为:' + str(asks_u))
    # print('档数为:' + str(len(asks_u)))
    # asks合并
    for i in asks_u:
        ask_price = i[0]
        for j in asks_p:
            if ask_price == j[0]:
                if i[1] == '0':
                    asks_p.remove(j)
                    break
                else:
                    del j[1]
                    j.insert(1, i[1])
                    break
        else:
            if i[1] != "0":
                asks_p.append(i)
    else:
        asks_p.sort(key=lambda price: sort_num(price[0]))
        # print('合并后的asks为:' + str(asks_p) + '，档数为:' + str(len(asks_p)))
    return asks_p


def sort_num(n):
    if n.isdigit():
        return int(n)
    else:
        return float(n)


def check(bids, asks):
    # 获取bid档str
    bids_l = []
    bid_l = []
    count_bid = 1
    while count_bid <= 25:
        if count_bid > len(bids):
            break
        bids_l.append(bids[count_bid - 1])
        count_bid += 1
    for j in bids_l:
        str_bid = ':'.join(j[0: 2])
        bid_l.append(str_bid)
    # 获取ask档str
    asks_l = []
    ask_l = []
    count_ask = 1
    while count_ask <= 25:
        if count_ask > len(asks):
            break
        asks_l.append(asks[count_ask - 1])
        count_ask += 1
    for k in asks_l:
        str_ask = ':'.join(k[0: 2])
        ask_l.append(str_ask)
    # 拼接str
    num = ''
    if len(bid_l) == len(ask_l):
        for m in range(len(bid_l)):
            num += bid_l[m] + ':' + ask_l[m] + ':'
    elif len(bid_l) > len(ask_l):
        # bid档比ask档多
        for n in range(len(ask_l)):
            num += bid_l[n] + ':' + ask_l[n] + ':'
        for l in range(len(ask_l), len(bid_l)):
            num += bid_l[l] + ':'
    elif len(bid_l) < len(ask_l):
        # ask档比bid档多
        for n in range(len(bid_l)):
            num += bid_l[n] + ':' + ask_l[n] + ':'
        for l in range(len(bid_l), len(ask_l)):
            num += ask_l[l] + ':'

    new_num = num[:-1]
    int_checksum = zlib.crc32(new_num.encode())
    fina = change(int_checksum)
    return fina


def change(num_old):
    num = pow(2, 31) - 1
    if num_old > num:
        out = num_old - num * 2 - 2
    else:
        out = num_old
    return out



import logging
import os
from datetime import datetime, timedelta

from ccxt import Exchange
from notecoin.coins.base.load import LoadDataKline
from notefile.compress import tarfile

logger = logging.getLogger()


def load_month(exchange: Exchange, tmp_path, timeframe='1m'):
    exchan = LoadDataKline(exchange)
    for i in range(1, 10):
        today = datetime.today()
        month_index = 1

        unix_start = datetime(today.year, today.month - month_index, 1).timestamp() * 1000
        unix_end = (datetime(today.year, today.month + 1 - month_index, 1) - timedelta(seconds=1)).timestamp() * 1000
        print(unix_start, unix_end)

        filename = f"{exchange.name}-kline-{datetime(today.year, today.month - month_index, 1).strftime('%Y-%m')}"
        path_csv = f"{tmp_path}/kline/{filename}.csv"
        path_tar = f"{tmp_path}/kline/{filename}.tar.xz"
        if os.path.exists(path_tar):
            continue

        # 下载
        exchan.load_all(timeframe=timeframe)
        # 保存
        exchan.table.to_csv(path_tar)
        # 压缩
        with tarfile.open(path_tar, "w:xz") as tar:
            tar.add(path_csv)
        # 删除
        os.remove(path_csv)
        exchan.table.delete_all()



import os

import ccxt
from notecoin.coins.base.cron import load_month
from notefile.compress import tarfile

path_root = '/home/bingtao/workspace/tmp/notecoin/'

# with tarfile.open(path_tar, "w:xz") as tar:
#     tar.add(path_csv)
load_month(ccxt.okex(),tmp_path=path_root)
